/**
 * Copyright (c) 2004-2015, Istvan David, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 * Istvan David - initial API and implementation
 */
package org.eclipse.viatra.cep.core.engine.compiler.rules;

import com.google.common.base.Function;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableListMultimap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Multimaps;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.viatra.cep.core.engine.compiler.EpsilonTransitionMatch;
import org.eclipse.viatra.cep.core.engine.compiler.EpsilonTransitionMatcher;
import org.eclipse.viatra.cep.core.engine.compiler.EquivalentStatesMatch;
import org.eclipse.viatra.cep.core.engine.compiler.EquivalentStatesMatcher;
import org.eclipse.viatra.cep.core.engine.compiler.EquivalentTransitionsMatch;
import org.eclipse.viatra.cep.core.engine.compiler.EquivalentTransitionsMatcher;
import org.eclipse.viatra.cep.core.engine.compiler.TransformationBasedCompiler;
import org.eclipse.viatra.cep.core.engine.compiler.builders.BuilderPrimitives;
import org.eclipse.viatra.cep.core.engine.compiler.rules.MappingRules;
import org.eclipse.viatra.cep.core.metamodels.automaton.Automaton;
import org.eclipse.viatra.cep.core.metamodels.automaton.EpsilonTransition;
import org.eclipse.viatra.cep.core.metamodels.automaton.FinalState;
import org.eclipse.viatra.cep.core.metamodels.automaton.Guard;
import org.eclipse.viatra.cep.core.metamodels.automaton.InitState;
import org.eclipse.viatra.cep.core.metamodels.automaton.InternalModel;
import org.eclipse.viatra.cep.core.metamodels.automaton.Parameter;
import org.eclipse.viatra.cep.core.metamodels.automaton.State;
import org.eclipse.viatra.cep.core.metamodels.automaton.TimedZone;
import org.eclipse.viatra.cep.core.metamodels.automaton.Transition;
import org.eclipse.viatra.cep.core.metamodels.automaton.TypedTransition;
import org.eclipse.viatra.cep.core.metamodels.events.EventPattern;
import org.eclipse.viatra.cep.core.metamodels.trace.TraceModel;
import org.eclipse.viatra.query.runtime.api.IMatchProcessor;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.impl.BaseMatcher;
import org.eclipse.viatra.query.runtime.api.impl.BasePatternMatch;
import org.eclipse.viatra.transformation.runtime.emf.rules.batch.BatchTransformationRule;
import org.eclipse.viatra.transformation.runtime.emf.rules.batch.BatchTransformationRuleFactory;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function0;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

@SuppressWarnings("all")
public class OptimizationRules extends MappingRules {
  @Extension
  private BuilderPrimitives builderPrimitives;
  
  public OptimizationRules(final InternalModel internalModel, final TraceModel traceModel) {
    super(internalModel, traceModel);
    BuilderPrimitives _builderPrimitives = new BuilderPrimitives(traceModel);
    this.builderPrimitives = _builderPrimitives;
  }
  
  @Override
  public List<? extends BatchTransformationRule<? extends BasePatternMatch, ? extends BaseMatcher<? extends BasePatternMatch>>> getAllRules() {
    return Collections.<BatchTransformationRule<? extends BasePatternMatch, ? extends BaseMatcher<? extends BasePatternMatch>>>unmodifiableList(CollectionLiterals.<BatchTransformationRule<? extends BasePatternMatch, ? extends BaseMatcher<? extends BasePatternMatch>>>newArrayList(this.mergeUponEpsilonTransitionRule, this.mergeEquivalentTransitionsRule, this.mergeEquivalentStatesRule));
  }
  
  /**
   * Transformation rule to merge {@link State}s connected by an {@link EpsilonTransition}.
   */
  private final BatchTransformationRule<EpsilonTransitionMatch, EpsilonTransitionMatcher> mergeUponEpsilonTransitionRule = new Function0<BatchTransformationRule<EpsilonTransitionMatch, EpsilonTransitionMatcher>>() {
    public BatchTransformationRule<EpsilonTransitionMatch, EpsilonTransitionMatcher> apply() {
      try {
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EpsilonTransitionMatch, EpsilonTransitionMatcher> _createRule = OptimizationRules.this.ruleFactory.<EpsilonTransitionMatch, EpsilonTransitionMatcher>createRule();
        IQuerySpecification<EpsilonTransitionMatcher> _querySpecification = EpsilonTransitionMatcher.querySpecification();
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EpsilonTransitionMatch, EpsilonTransitionMatcher> _precondition = _createRule.precondition(_querySpecification);
        final IMatchProcessor<EpsilonTransitionMatch> _function = new IMatchProcessor<EpsilonTransitionMatch>() {
          @Override
          public void process(final EpsilonTransitionMatch it) {
            State _preState = it.getPreState();
            State _postState = it.getPostState();
            boolean _equals = Objects.equal(_preState, _postState);
            if (_equals) {
              EpsilonTransition _transition = it.getTransition();
              OptimizationRules.this.builderPrimitives.removeTransition(_transition);
            } else {
              boolean _or = false;
              State _preState_1 = it.getPreState();
              if ((_preState_1 instanceof InitState)) {
                _or = true;
              } else {
                State _preState_2 = it.getPreState();
                _or = (_preState_2 instanceof FinalState);
              }
              final boolean keepPreState = _or;
              boolean _or_1 = false;
              State _postState_1 = it.getPostState();
              if ((_postState_1 instanceof InitState)) {
                _or_1 = true;
              } else {
                State _postState_2 = it.getPostState();
                _or_1 = (_postState_2 instanceof FinalState);
              }
              final boolean keepPostState = _or_1;
              Preconditions.checkArgument((!(keepPreState && keepPostState)));
              if (keepPreState) {
                State _postState_3 = it.getPostState();
                State _preState_3 = it.getPreState();
                EpsilonTransition _transition_1 = it.getTransition();
                OptimizationRules.this.mergeStates(_postState_3, _preState_3, _transition_1);
              } else {
                if (keepPostState) {
                  State _preState_4 = it.getPreState();
                  State _postState_4 = it.getPostState();
                  EpsilonTransition _transition_2 = it.getTransition();
                  OptimizationRules.this.mergeStates(_preState_4, _postState_4, _transition_2);
                } else {
                  State _preState_5 = it.getPreState();
                  State _postState_5 = it.getPostState();
                  EpsilonTransition _transition_3 = it.getTransition();
                  OptimizationRules.this.mergeStates(_preState_5, _postState_5, _transition_3);
                }
              }
            }
          }
        };
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EpsilonTransitionMatch, EpsilonTransitionMatcher> _action = _precondition.action(_function);
        BatchTransformationRule<EpsilonTransitionMatch, EpsilonTransitionMatcher> _build = _action.build();
        return _build;
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    }
  }.apply();
  
  /**
   * Transformation rule to merge equivalent {@link Transition}s.
   */
  private final BatchTransformationRule<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> mergeEquivalentTransitionsRule = new Function0<BatchTransformationRule<EquivalentTransitionsMatch, EquivalentTransitionsMatcher>>() {
    public BatchTransformationRule<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> apply() {
      try {
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> _createRule = OptimizationRules.this.ruleFactory.<EquivalentTransitionsMatch, EquivalentTransitionsMatcher>createRule();
        IQuerySpecification<EquivalentTransitionsMatcher> _querySpecification = EquivalentTransitionsMatcher.querySpecification();
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> _precondition = _createRule.precondition(_querySpecification);
        final IMatchProcessor<EquivalentTransitionsMatch> _function = new IMatchProcessor<EquivalentTransitionsMatch>() {
          @Override
          public void process(final EquivalentTransitionsMatch it) {
            TypedTransition _transition2 = it.getTransition2();
            TypedTransition _transition1 = it.getTransition1();
            OptimizationRules.this.unifyParameters(_transition2, _transition1);
            TypedTransition _transition1_1 = it.getTransition1();
            OptimizationRules.this.builderPrimitives.removeTransition(_transition1_1);
          }
        };
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> _action = _precondition.action(_function);
        BatchTransformationRule<EquivalentTransitionsMatch, EquivalentTransitionsMatcher> _build = _action.build();
        return _build;
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    }
  }.apply();
  
  /**
   * Transformation rule to merge equivalent {@link State}s.
   */
  private final BatchTransformationRule<EquivalentStatesMatch, EquivalentStatesMatcher> mergeEquivalentStatesRule = new Function0<BatchTransformationRule<EquivalentStatesMatch, EquivalentStatesMatcher>>() {
    public BatchTransformationRule<EquivalentStatesMatch, EquivalentStatesMatcher> apply() {
      try {
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentStatesMatch, EquivalentStatesMatcher> _createRule = OptimizationRules.this.ruleFactory.<EquivalentStatesMatch, EquivalentStatesMatcher>createRule();
        IQuerySpecification<EquivalentStatesMatcher> _querySpecification = EquivalentStatesMatcher.querySpecification();
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentStatesMatch, EquivalentStatesMatcher> _precondition = _createRule.precondition(_querySpecification);
        final IMatchProcessor<EquivalentStatesMatch> _function = new IMatchProcessor<EquivalentStatesMatch>() {
          @Override
          public void process(final EquivalentStatesMatch it) {
            State _postState1 = it.getPostState1();
            State _postState2 = it.getPostState2();
            boolean _equals = Objects.equal(_postState1, _postState2);
            if (_equals) {
              State _postState2_1 = it.getPostState2();
              State _postState1_1 = it.getPostState1();
              TypedTransition _transition2 = it.getTransition2();
              TypedTransition _transition1 = it.getTransition1();
              OptimizationRules.this.mergeStates(_postState2_1, _postState1_1, _transition2, _transition1);
            } else {
              boolean _and = false;
              State _postState1_2 = it.getPostState1();
              if (!(_postState1_2 instanceof InitState)) {
                _and = false;
              } else {
                State _postState2_2 = it.getPostState2();
                _and = (_postState2_2 instanceof FinalState);
              }
              boolean _not = (!_and);
              Preconditions.checkArgument(_not);
              boolean _and_1 = false;
              State _postState1_3 = it.getPostState1();
              if (!(_postState1_3 instanceof FinalState)) {
                _and_1 = false;
              } else {
                State _postState2_3 = it.getPostState2();
                _and_1 = (_postState2_3 instanceof InitState);
              }
              boolean _not_1 = (!_and_1);
              Preconditions.checkArgument(_not_1);
              State _postState1_4 = it.getPostState1();
              boolean _matched = false;
              if (!_matched) {
                if (_postState1_4 instanceof InitState) {
                  _matched=true;
                  State _postState2_4 = it.getPostState2();
                  State _postState1_5 = it.getPostState1();
                  TypedTransition _transition2_1 = it.getTransition2();
                  TypedTransition _transition1_1 = it.getTransition1();
                  OptimizationRules.this.mergeStates(_postState2_4, _postState1_5, _transition2_1, _transition1_1);
                }
              }
              if (!_matched) {
                if (_postState1_4 instanceof FinalState) {
                  _matched=true;
                  State _postState2_4 = it.getPostState2();
                  State _postState1_5 = it.getPostState1();
                  TypedTransition _transition2_1 = it.getTransition2();
                  TypedTransition _transition1_1 = it.getTransition1();
                  OptimizationRules.this.mergeStates(_postState2_4, _postState1_5, _transition2_1, _transition1_1);
                }
              }
              if (!_matched) {
                State _postState2_4 = it.getPostState2();
                boolean _matched_1 = false;
                if (!_matched_1) {
                  if (_postState2_4 instanceof InitState) {
                    _matched_1=true;
                    State _postState1_5 = it.getPostState1();
                    State _postState2_5 = it.getPostState2();
                    TypedTransition _transition1_1 = it.getTransition1();
                    TypedTransition _transition2_1 = it.getTransition2();
                    OptimizationRules.this.mergeStates(_postState1_5, _postState2_5, _transition1_1, _transition2_1);
                  }
                }
                if (!_matched_1) {
                  if (_postState2_4 instanceof FinalState) {
                    _matched_1=true;
                    State _postState1_5 = it.getPostState1();
                    State _postState2_5 = it.getPostState2();
                    TypedTransition _transition1_1 = it.getTransition1();
                    TypedTransition _transition2_1 = it.getTransition2();
                    OptimizationRules.this.mergeStates(_postState1_5, _postState2_5, _transition1_1, _transition2_1);
                  }
                }
                if (!_matched_1) {
                  State _postState1_5 = it.getPostState1();
                  State _postState2_5 = it.getPostState2();
                  TypedTransition _transition1_1 = it.getTransition1();
                  TypedTransition _transition2_1 = it.getTransition2();
                  OptimizationRules.this.mergeStates(_postState1_5, _postState2_5, _transition1_1, _transition2_1);
                }
              }
            }
          }
        };
        BatchTransformationRuleFactory.BatchTransformationRuleBuilder<EquivalentStatesMatch, EquivalentStatesMatcher> _action = _precondition.action(_function);
        BatchTransformationRule<EquivalentStatesMatch, EquivalentStatesMatcher> _build = _action.build();
        return _build;
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    }
  }.apply();
  
  /**
   * Merge two {@link State}s with the respective associated {@link Transition}s.
   */
  private boolean mergeStates(final State stateToDelete, final State stateToKeep, final TypedTransition transitionToRemove, final TypedTransition transitionToKeep) {
    boolean _xblockexpression = false;
    {
      this.unifyParameters(transitionToKeep, transitionToRemove);
      _xblockexpression = this.mergeStates(stateToDelete, stateToKeep, transitionToRemove);
    }
    return _xblockexpression;
  }
  
  /**
   * Merge two {@link State}s through a {@link Transition}.
   */
  private boolean mergeStates(final State stateToDelete, final State stateToKeep, final Transition transition) {
    boolean _xblockexpression = false;
    {
      this.builderPrimitives.removeTransition(transition);
      boolean _xifexpression = false;
      boolean _notEquals = (!Objects.equal(stateToDelete, stateToKeep));
      if (_notEquals) {
        _xifexpression = this.mergeStates(stateToDelete, stateToKeep);
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  /**
   * Merge two {@link State}s.
   */
  private boolean mergeStates(final State stateToDelete, final State stateToKeep) {
    boolean _xblockexpression = false;
    {
      EList<Transition> _inTransitions = stateToKeep.getInTransitions();
      EList<Transition> _inTransitions_1 = stateToDelete.getInTransitions();
      Iterables.<Transition>addAll(_inTransitions, _inTransitions_1);
      EList<Transition> _outTransitions = stateToKeep.getOutTransitions();
      EList<Transition> _outTransitions_1 = stateToDelete.getOutTransitions();
      Iterables.<Transition>addAll(_outTransitions, _outTransitions_1);
      EList<TimedZone> _inStateOf = stateToKeep.getInStateOf();
      EList<TimedZone> _inStateOf_1 = stateToDelete.getInStateOf();
      Iterables.<TimedZone>addAll(_inStateOf, _inStateOf_1);
      EList<TimedZone> _outStateOf = stateToKeep.getOutStateOf();
      EList<TimedZone> _outStateOf_1 = stateToDelete.getOutStateOf();
      Iterables.<TimedZone>addAll(_outStateOf, _outStateOf_1);
      EList<Transition> _outTransitions_2 = stateToDelete.getOutTransitions();
      final Function1<Transition, Boolean> _function = new Function1<Transition, Boolean>() {
        @Override
        public Boolean apply(final Transition t) {
          return Boolean.valueOf((t instanceof EpsilonTransition));
        }
      };
      boolean _forall = IterableExtensions.<Transition>forall(_outTransitions_2, _function);
      Preconditions.checkArgument(_forall);
      _xblockexpression = this.builderPrimitives.removeState(stateToDelete);
    }
    return _xblockexpression;
  }
  
  /**
   * Unify the parameters of two {@link Transitions}s.
   */
  protected void unifyParameters(final TypedTransition transitionToKeep, final TypedTransition transitionToRemove) {
    EList<Guard> _guards = transitionToKeep.getGuards();
    final Function1<Guard, String> _function = new Function1<Guard, String>() {
      @Override
      public String apply(final Guard it) {
        EventPattern _eventType = it.getEventType();
        return _eventType.getId();
      }
    };
    List<String> _map = ListExtensions.<Guard, String>map(_guards, _function);
    EList<Guard> _guards_1 = transitionToRemove.getGuards();
    final Function1<Guard, String> _function_1 = new Function1<Guard, String>() {
      @Override
      public String apply(final Guard it) {
        EventPattern _eventType = it.getEventType();
        return _eventType.getId();
      }
    };
    List<String> _map_1 = ListExtensions.<Guard, String>map(_guards_1, _function_1);
    boolean _equals = Objects.equal(_map, _map_1);
    Preconditions.checkArgument(_equals);
    EObject _eContainer = transitionToKeep.eContainer();
    EObject _eContainer_1 = _eContainer.eContainer();
    final Automaton automaton = ((Automaton) _eContainer_1);
    final Collection<Runnable> unifications = CollectionLiterals.<Runnable>newArrayList();
    EList<Parameter> _parameters = transitionToKeep.getParameters();
    final Function<Parameter, Integer> _function_2 = new Function<Parameter, Integer>() {
      @Override
      public Integer apply(final Parameter it) {
        return Integer.valueOf(it.getPosition());
      }
    };
    final ImmutableListMultimap<Integer, Parameter> toKeepParamsByPosition = Multimaps.<Integer, Parameter>index(_parameters, _function_2);
    EList<Parameter> _parameters_1 = transitionToRemove.getParameters();
    for (final Parameter b : _parameters_1) {
      {
        final String bName = b.getSymbolicName();
        int _position = b.getPosition();
        final ImmutableList<Parameter> toKeepParams = toKeepParamsByPosition.get(Integer.valueOf(_position));
        boolean _isEmpty = toKeepParams.isEmpty();
        if (_isEmpty) {
          final Function1<Object, Boolean> _function_3 = new Function1<Object, Boolean>() {
            @Override
            public Boolean apply(final Object it) {
              EList<Parameter> _parameters = transitionToKeep.getParameters();
              return Boolean.valueOf(_parameters.add(b));
            }
          };
          unifications.add(((Runnable) new Runnable() {
              public void run() {
                _function_3.apply(null);
              }
          }));
        } else {
          for (final Parameter a : toKeepParams) {
            {
              final String aName = a.getSymbolicName();
              boolean _notEquals = (!Objects.equal(aName, bName));
              if (_notEquals) {
                boolean _equals_1 = Objects.equal(bName, TransformationBasedCompiler.OMITTED_PARAMETER_SYMBOLIC_NAME);
                if (_equals_1) {
                } else {
                  boolean _equals_2 = Objects.equal(aName, TransformationBasedCompiler.OMITTED_PARAMETER_SYMBOLIC_NAME);
                  if (_equals_2) {
                    Runnable _localUnification = this.getLocalUnification(transitionToKeep, b, a);
                    unifications.add(_localUnification);
                  } else {
                    boolean _contains = aName.contains("|");
                    if (_contains) {
                      Runnable _globalUnification = this.getGlobalUnification(automaton, bName, aName);
                      unifications.add(_globalUnification);
                    } else {
                      boolean _contains_1 = bName.contains("|");
                      Preconditions.checkState(_contains_1);
                      Runnable _globalUnification_1 = this.getGlobalUnification(automaton, aName, bName);
                      unifications.add(_globalUnification_1);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    final Procedure1<Runnable> _function_3 = new Procedure1<Runnable>() {
      @Override
      public void apply(final Runnable it) {
        it.run();
      }
    };
    IterableExtensions.<Runnable>forEach(unifications, _function_3);
  }
  
  public Runnable getLocalUnification(final TypedTransition transition, final Parameter toKeep, final Parameter toRemove) {
    final Runnable _function = new Runnable() {
      @Override
      public void run() {
        EList<Parameter> _parameters = transition.getParameters();
        _parameters.remove(toRemove);
        EList<Parameter> _parameters_1 = transition.getParameters();
        _parameters_1.add(toKeep);
      }
    };
    return _function;
  }
  
  public Runnable getGlobalUnification(final Automaton automaton, final String keepName, final String removeName) {
    final Runnable _function = new Runnable() {
      @Override
      public void run() {
        EList<State> _states = automaton.getStates();
        final Procedure1<State> _function = new Procedure1<State>() {
          @Override
          public void apply(final State it) {
            EList<Transition> _outTransitions = it.getOutTransitions();
            Iterable<TypedTransition> _filter = Iterables.<TypedTransition>filter(_outTransitions, TypedTransition.class);
            final Procedure1<TypedTransition> _function = new Procedure1<TypedTransition>() {
              @Override
              public void apply(final TypedTransition it) {
                EList<Parameter> _parameters = it.getParameters();
                final Procedure1<Parameter> _function = new Procedure1<Parameter>() {
                  @Override
                  public void apply(final Parameter it) {
                    String _symbolicName = it.getSymbolicName();
                    boolean _equals = Objects.equal(_symbolicName, removeName);
                    if (_equals) {
                      it.setSymbolicName(keepName);
                    }
                  }
                };
                IterableExtensions.<Parameter>forEach(_parameters, _function);
              }
            };
            IterableExtensions.<TypedTransition>forEach(_filter, _function);
          }
        };
        IterableExtensions.<State>forEach(_states, _function);
      }
    };
    return _function;
  }
}
