/**
 * Copyright (c) 2010-2017, Grill Balázs, IncQueryLabs
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Balázs - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.coverage;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import org.eclipse.viatra.query.runtime.matchers.psystem.PBody;
import org.eclipse.viatra.query.runtime.matchers.psystem.PTraceable;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PDisjunction;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQuery;
import org.eclipse.viatra.query.testing.core.coverage.CoverageContext;
import org.eclipse.viatra.query.testing.core.coverage.CoverageState;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * Associates {@link CoverageState}s to objects.
 * @param <T> type of objects whose coverages are stored
 * @since 1.6
 */
@SuppressWarnings("all")
public class CoverageInfo<T extends Object> extends HashMap<CoverageContext<T>, CoverageState> {
  protected final static long serialVersionUID = (-8699692647123679741L);
  
  /**
   * Merge coverage. The keys of this and other should be disjunct.
   */
  public CoverageInfo<T> mergeWith(final CoverageInfo<T> other) {
    final CoverageInfo<T> result = new CoverageInfo<T>();
    result.putAll(this);
    result.putAll(other);
    return result;
  }
  
  /**
   * Returns the coverage for elements aggregated per scope.
   */
  public Map<T, CoverageState> getElementCoverage() {
    Map<T, CoverageState> _xblockexpression = null;
    {
      final Map<T, CoverageState> result = CollectionLiterals.<T, CoverageState>newHashMap();
      Set<Map.Entry<CoverageContext<T>, CoverageState>> _entrySet = this.entrySet();
      for (final Map.Entry<CoverageContext<T>, CoverageState> it : _entrySet) {
        {
          CoverageContext<T> _key = it.getKey();
          final T element = _key.getElement();
          final CoverageState state = it.getValue();
          final CoverageState otherState = result.get(element);
          if (((state != null) && (otherState != null))) {
            CoverageState _best = state.best(otherState);
            result.put(element, _best);
          } else {
            CoverageState _elvis = null;
            if (state != null) {
              _elvis = state;
            } else {
              _elvis = otherState;
            }
            result.put(element, _elvis);
          }
        }
      }
      _xblockexpression = result;
    }
    return _xblockexpression;
  }
  
  /**
   * Returns the ratio of the covered and the represented {@link PBody}s contained in this info, in percents.
   */
  public double getAggregatedCoveragePercent() {
    Map<T, CoverageState> _elementCoverage = this.getElementCoverage();
    Set<T> _keySet = _elementCoverage.keySet();
    Iterable<PBody> _filter = Iterables.<PBody>filter(_keySet, PBody.class);
    return this.getCoveragePercent(_filter);
  }
  
  /**
   * Returns the ratio of the given covered and the represented {@link PTraceable}s, in percents.
   */
  public double getCoveragePercent(final Iterable<? extends PTraceable> traceables) {
    double _xblockexpression = (double) 0;
    {
      final Map<T, CoverageState> elementCoverage = this.getElementCoverage();
      final Function1<PTraceable, Boolean> _function = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          CoverageState _get = elementCoverage.get(it);
          return Boolean.valueOf(Objects.equal(_get, CoverageState.COVERED));
        }
      };
      Iterable<? extends PTraceable> _filter = IterableExtensions.filter(traceables, _function);
      int _size = IterableExtensions.size(_filter);
      final double coveredCount = ((double) _size);
      final Function1<PTraceable, Boolean> _function_1 = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          CoverageState _get = elementCoverage.get(it);
          return Boolean.valueOf(Collections.<CoverageState>unmodifiableSet(CollectionLiterals.<CoverageState>newHashSet(CoverageState.COVERED, CoverageState.NOT_COVERED)).contains(_get));
        }
      };
      Iterable<? extends PTraceable> _filter_1 = IterableExtensions.filter(traceables, _function_1);
      final int representedCount = IterableExtensions.size(_filter_1);
      double _xifexpression = (double) 0;
      if ((representedCount == 0)) {
        _xifexpression = 0;
      } else {
        _xifexpression = ((coveredCount / representedCount) * 100);
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  /**
   * Returns the ratio of the covered and the represented {@link PBody}s contained in the given {@link PQuery}, in percents.
   */
  public double getCoveragePercent(final PQuery query) {
    PDisjunction _disjunctBodies = query.getDisjunctBodies();
    Set<PBody> _bodies = _disjunctBodies.getBodies();
    return this.getCoveragePercent(_bodies);
  }
}
