/**
 * Copyright (c) 2010-2017, Dénes Harmath, IncQueryLabs
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Dénes Harmath - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.coverage;

import com.google.common.base.Charsets;
import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.common.io.Files;
import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.emf.common.util.URI;
import org.eclipse.viatra.query.runtime.api.impl.BaseGeneratedPrivateEMFQuerySpecification;
import org.eclipse.viatra.query.runtime.matchers.psystem.PBody;
import org.eclipse.viatra.query.runtime.matchers.psystem.PConstraint;
import org.eclipse.viatra.query.runtime.matchers.psystem.PTraceable;
import org.eclipse.viatra.query.runtime.matchers.psystem.basicdeferred.ExportedParameter;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PDisjunction;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQuery;
import org.eclipse.viatra.query.testing.core.coverage.CoverageAnalyzer;
import org.eclipse.viatra.query.testing.core.coverage.CoverageContext;
import org.eclipse.viatra.query.testing.core.coverage.CoverageInfo;
import org.eclipse.viatra.query.testing.core.coverage.CoverageState;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.InputOutput;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * Utility methods to report the results of a {@link CoverageAnalyzer}.
 * 
 * @since 1.6
 */
@SuppressWarnings("all")
public class CoverageReporter {
  /**
   * Prints coverage report to the standard output.
   */
  public static void reportConsole(final CoverageAnalyzer it) {
    StringConcatenation _builder = new StringConcatenation();
    {
      CoverageInfo<PTraceable> _coverage = it.getCoverage();
      Map<PTraceable, CoverageState> _elementCoverage = _coverage.getElementCoverage();
      Set<PTraceable> _keySet = _elementCoverage.keySet();
      Iterable<PQuery> _filter = Iterables.<PQuery>filter(_keySet, PQuery.class);
      final Function1<PQuery, String> _function = new Function1<PQuery, String>() {
        @Override
        public String apply(final PQuery it) {
          return it.getFullyQualifiedName();
        }
      };
      List<PQuery> _sortBy = IterableExtensions.<PQuery, String>sortBy(_filter, _function);
      for(final PQuery query : _sortBy) {
        _builder.append("pattern ");
        String _fullyQualifiedName = query.getFullyQualifiedName();
        _builder.append(_fullyQualifiedName, "");
        _builder.append(" (");
        CoverageInfo<PTraceable> _coverage_1 = it.getCoverage();
        Map<PTraceable, CoverageState> _elementCoverage_1 = _coverage_1.getElementCoverage();
        CoverageState _get = _elementCoverage_1.get(query);
        _builder.append(_get, "");
        _builder.append(") ");
        CoverageInfo<PTraceable> _coverage_2 = it.getCoverage();
        double _coveragePercent = _coverage_2.getCoveragePercent(query);
        CharSequence _formatPercent = CoverageReporter.formatPercent(_coveragePercent);
        _builder.append(_formatPercent, "");
        _builder.newLineIfNotEmpty();
        {
          PDisjunction _disjunctBodies = query.getDisjunctBodies();
          Set<PBody> _bodies = _disjunctBodies.getBodies();
          boolean _hasElements = false;
          for(final PBody body : _bodies) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(" or ", "");
            }
            _builder.append("{ (");
            CoverageInfo<PTraceable> _coverage_3 = it.getCoverage();
            Map<PTraceable, CoverageState> _elementCoverage_2 = _coverage_3.getElementCoverage();
            CoverageState _get_1 = _elementCoverage_2.get(body);
            _builder.append(_get_1, "");
            _builder.append(")");
            _builder.newLineIfNotEmpty();
            {
              Set<PConstraint> _constraints = body.getConstraints();
              for(final PConstraint constraint : _constraints) {
                _builder.append("\t");
                _builder.append(constraint, "\t");
                _builder.append(" (");
                CoverageInfo<PTraceable> _coverage_4 = it.getCoverage();
                Map<PTraceable, CoverageState> _elementCoverage_3 = _coverage_4.getElementCoverage();
                CoverageState _get_2 = _elementCoverage_3.get(constraint);
                _builder.append(_get_2, "\t");
                _builder.append(")");
                _builder.newLineIfNotEmpty();
              }
            }
            _builder.append("}");
            _builder.newLine();
          }
        }
      }
    }
    InputOutput.<String>println(_builder.toString());
  }
  
  private static CharSequence formatPercent(final double percent) {
    StringConcatenation _builder = new StringConcatenation();
    String _format = String.format("%1$.2f", Double.valueOf(percent));
    _builder.append(_format, "");
    _builder.append("%");
    return _builder;
  }
  
  /**
   * Saves coverage report in HTML format to the given {@link File}.
   */
  public static void reportHtml(final CoverageAnalyzer it, final File file) {
    try {
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("<html><body>");
      _builder.newLine();
      _builder.append("<h1>Pattern coverage report</h1>");
      _builder.newLine();
      _builder.append("<h2>Pattern & body coverage</h2>");
      _builder.newLine();
      _builder.append("<h3>Overall: ");
      CoverageInfo<PTraceable> _coverage = it.getCoverage();
      CoverageInfo<PTraceable> _coverage_1 = it.getCoverage();
      Map<PTraceable, CoverageState> _elementCoverage = _coverage_1.getElementCoverage();
      Set<PTraceable> _keySet = _elementCoverage.keySet();
      Iterable<PBody> _filter = Iterables.<PBody>filter(_keySet, PBody.class);
      CharSequence _bodyCoverage = CoverageReporter.getBodyCoverage(_coverage, _filter);
      _builder.append(_bodyCoverage, "");
      _builder.append("</h3>");
      _builder.newLineIfNotEmpty();
      _builder.append("<ul>");
      _builder.newLine();
      {
        CoverageInfo<PTraceable> _coverage_2 = it.getCoverage();
        Map<PTraceable, CoverageState> _elementCoverage_1 = _coverage_2.getElementCoverage();
        Set<PTraceable> _keySet_1 = _elementCoverage_1.keySet();
        Iterable<PQuery> _filter_1 = Iterables.<PQuery>filter(_keySet_1, PQuery.class);
        final Function1<PQuery, String> _function = new Function1<PQuery, String>() {
          @Override
          public String apply(final PQuery it) {
            return it.getFullyQualifiedName();
          }
        };
        List<PQuery> _sortBy = IterableExtensions.<PQuery, String>sortBy(_filter_1, _function);
        for(final PQuery query : _sortBy) {
          _builder.append("            ");
          PDisjunction _disjunctBodies = query.getDisjunctBodies();
          final Set<PBody> bodies = _disjunctBodies.getBodies();
          _builder.newLineIfNotEmpty();
          _builder.append("<li>");
          _builder.newLine();
          _builder.append("  ");
          _builder.append("<a href=\"#");
          String _fullyQualifiedName = query.getFullyQualifiedName();
          _builder.append(_fullyQualifiedName, "  ");
          _builder.append("\"><span ");
          CoverageInfo<PTraceable> _coverage_3 = it.getCoverage();
          String _attributes = CoverageReporter.getAttributes(_coverage_3, query, bodies);
          _builder.append(_attributes, "  ");
          _builder.append(">");
          _builder.newLineIfNotEmpty();
          _builder.append("  ");
          {
            List<Object> _publishedAs = query.publishedAs();
            Iterable<BaseGeneratedPrivateEMFQuerySpecification> _filter_2 = Iterables.<BaseGeneratedPrivateEMFQuerySpecification>filter(_publishedAs, BaseGeneratedPrivateEMFQuerySpecification.class);
            boolean _isEmpty = IterableExtensions.isEmpty(_filter_2);
            boolean _not = (!_isEmpty);
            if (_not) {
              _builder.append("private");
            }
          }
          _builder.newLineIfNotEmpty();
          _builder.append("  ");
          _builder.append("pattern ");
          String _fullyQualifiedName_1 = query.getFullyQualifiedName();
          _builder.append(_fullyQualifiedName_1, "  ");
          _builder.append("</span></a> ");
          CoverageInfo<PTraceable> _coverage_4 = it.getCoverage();
          CharSequence _bodyCoverage_1 = CoverageReporter.getBodyCoverage(_coverage_4, bodies);
          _builder.append(_bodyCoverage_1, "  ");
          _builder.newLineIfNotEmpty();
          _builder.append("</li>");
          _builder.newLine();
        }
      }
      _builder.append("</ul>");
      _builder.newLine();
      _builder.append("<h2>Detailed coverage</h2>");
      _builder.newLine();
      {
        CoverageInfo<PTraceable> _coverage_5 = it.getCoverage();
        Map<PTraceable, CoverageState> _elementCoverage_2 = _coverage_5.getElementCoverage();
        Set<PTraceable> _keySet_2 = _elementCoverage_2.keySet();
        Iterable<PQuery> _filter_3 = Iterables.<PQuery>filter(_keySet_2, PQuery.class);
        final Function1<PQuery, String> _function_1 = new Function1<PQuery, String>() {
          @Override
          public String apply(final PQuery it) {
            return it.getFullyQualifiedName();
          }
        };
        List<PQuery> _sortBy_1 = IterableExtensions.<PQuery, String>sortBy(_filter_3, _function_1);
        for(final PQuery query_1 : _sortBy_1) {
          _builder.append("<h3 id=\"");
          String _fullyQualifiedName_2 = query_1.getFullyQualifiedName();
          _builder.append(_fullyQualifiedName_2, "");
          _builder.append("\"/>pattern ");
          String _fullyQualifiedName_3 = query_1.getFullyQualifiedName();
          _builder.append(_fullyQualifiedName_3, "");
          _builder.append("</h3>");
          _builder.newLineIfNotEmpty();
          {
            CoverageInfo<PTraceable> _coverage_6 = it.getCoverage();
            Set<CoverageContext<PTraceable>> _keySet_3 = _coverage_6.keySet();
            final Function1<CoverageContext<PTraceable>, Boolean> _function_2 = new Function1<CoverageContext<PTraceable>, Boolean>() {
              @Override
              public Boolean apply(final CoverageContext<PTraceable> it) {
                PTraceable _element = it.getElement();
                return Boolean.valueOf(Objects.equal(_element, query_1));
              }
            };
            Iterable<CoverageContext<PTraceable>> _filter_4 = IterableExtensions.<CoverageContext<PTraceable>>filter(_keySet_3, _function_2);
            for(final CoverageContext<PTraceable> context : _filter_4) {
              _builder.append("<p>Model: ");
              Set<URI> _scope = context.getScope();
              _builder.append(_scope, "");
              _builder.append("</p>");
              _builder.newLineIfNotEmpty();
              _builder.append("<p>");
              _builder.newLine();
              int bodyIndex = 0;
              _builder.newLineIfNotEmpty();
              {
                PDisjunction _disjunctBodies_1 = query_1.getDisjunctBodies();
                Set<PBody> _bodies = _disjunctBodies_1.getBodies();
                boolean _hasElements = false;
                for(final PBody body : _bodies) {
                  if (!_hasElements) {
                    _hasElements = true;
                  } else {
                    _builder.appendImmediate(" or ", "             ");
                  }
                  _builder.append("             ");
                  _builder.append(" ");
                  Set<PConstraint> _constraints = body.getConstraints();
                  final Function1<PConstraint, Boolean> _function_3 = new Function1<PConstraint, Boolean>() {
                    @Override
                    public Boolean apply(final PConstraint it) {
                      return Boolean.valueOf((!(it instanceof ExportedParameter)));
                    }
                  };
                  final Iterable<PConstraint> constraints = IterableExtensions.<PConstraint>filter(_constraints, _function_3);
                  _builder.newLineIfNotEmpty();
                  _builder.append("             ");
                  _builder.append("<div><span ");
                  CoverageInfo<PTraceable> _coverage_7 = it.getCoverage();
                  Set<URI> _scope_1 = context.getScope();
                  String _attributes_1 = CoverageReporter.getAttributes(_coverage_7, body, constraints, _scope_1);
                  _builder.append(_attributes_1, "             ");
                  _builder.append(">Body #");
                  int _plusPlus = bodyIndex++;
                  _builder.append(_plusPlus, "             ");
                  _builder.append(" {</span>");
                  _builder.newLineIfNotEmpty();
                  {
                    for(final PConstraint constraint : constraints) {
                      _builder.append("             ");
                      _builder.append("<div ");
                      CoverageInfo<PTraceable> _coverage_8 = it.getCoverage();
                      Set<URI> _scope_2 = context.getScope();
                      String _attributes_2 = CoverageReporter.getAttributes(_coverage_8, constraint, Collections.<PTraceable>unmodifiableList(CollectionLiterals.<PTraceable>newArrayList()), _scope_2);
                      _builder.append(_attributes_2, "             ");
                      _builder.append("\">");
                      _builder.newLineIfNotEmpty();
                      _builder.append("             ");
                      _builder.append(constraint, "             ");
                      _builder.newLineIfNotEmpty();
                      _builder.append("             ");
                      _builder.append("</div>");
                      _builder.newLine();
                    }
                  }
                  _builder.append("             ");
                  _builder.append("}</div>");
                  _builder.newLine();
                }
              }
            }
          }
          _builder.append("</p>");
          _builder.newLine();
        }
      }
      _builder.append("<h2>Legend</h2>");
      _builder.newLine();
      _builder.append("<p>");
      _builder.newLine();
      _builder.append("\t");
      String _legend = CoverageReporter.getLegend(CoverageState.COVERED, true);
      _builder.append(_legend, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      String _legend_1 = CoverageReporter.getLegend(CoverageState.COVERED, false);
      _builder.append(_legend_1, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      String _legend_2 = CoverageReporter.getLegend(CoverageState.NOT_COVERED, false);
      _builder.append(_legend_2, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      String _legend_3 = CoverageReporter.getLegend(CoverageState.NOT_REPRESENTED, true);
      _builder.append(_legend_3, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      String _legend_4 = CoverageReporter.getLegend(CoverageState.NOT_REPRESENTED_UNKNOWN_REASON, true);
      _builder.append(_legend_4, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("</p>");
      _builder.newLine();
      _builder.append("</body></html>");
      final String content = _builder.toString();
      Files.write(content, file, Charsets.UTF_8);
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private static CharSequence getBodyCoverage(final CoverageInfo<?> coverage, final Iterable<PBody> bodies) {
    StringConcatenation _builder = new StringConcatenation();
    double _coveragePercent = coverage.getCoveragePercent(bodies);
    CharSequence _formatPercent = CoverageReporter.formatPercent(_coveragePercent);
    _builder.append(_formatPercent, "");
    _builder.newLineIfNotEmpty();
    _builder.append("        ");
    _builder.append("(");
    final Function1<PBody, Boolean> _function = new Function1<PBody, Boolean>() {
      @Override
      public Boolean apply(final PBody body) {
        Map<?, CoverageState> _elementCoverage = coverage.getElementCoverage();
        CoverageState _get = _elementCoverage.get(body);
        return Boolean.valueOf(Objects.equal(_get, CoverageState.COVERED));
      }
    };
    Iterable<PBody> _filter = IterableExtensions.<PBody>filter(bodies, _function);
    int _size = IterableExtensions.size(_filter);
    _builder.append(_size, "        ");
    _builder.append("/");
    int _size_1 = IterableExtensions.size(bodies);
    _builder.append(_size_1, "        ");
    _builder.append(" bodies covered)");
    return _builder;
  }
  
  private static String getAttributes(final CoverageInfo<?> coverage, final PTraceable traceable, final Iterable<? extends PTraceable> children) {
    String _xblockexpression = null;
    {
      CoverageState _elvis = null;
      Map<?, CoverageState> _elementCoverage = coverage.getElementCoverage();
      CoverageState _get = _elementCoverage.get(traceable);
      if (_get != null) {
        _elvis = _get;
      } else {
        _elvis = CoverageState.NOT_REPRESENTED_UNKNOWN_REASON;
      }
      final CoverageState state = _elvis;
      final Function1<PTraceable, Boolean> _function = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          Map<?, CoverageState> _elementCoverage = coverage.getElementCoverage();
          CoverageState _get = _elementCoverage.get(it);
          return Boolean.valueOf((!Objects.equal(_get, CoverageState.NOT_COVERED)));
        }
      };
      final boolean childrenCovered = IterableExtensions.forall(children, _function);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("style=\"");
      String _style = CoverageReporter.getStyle(state, childrenCovered);
      _builder.append(_style, "");
      _builder.append("\" title=\"");
      String _title = CoverageReporter.getTitle(state, childrenCovered);
      _builder.append(_title, "");
      _builder.append("\"");
      _xblockexpression = _builder.toString();
    }
    return _xblockexpression;
  }
  
  private static String getAttributes(final CoverageInfo<?> coverage, final PTraceable traceable, final Iterable<? extends PTraceable> children, final Set<URI> scope) {
    String _xblockexpression = null;
    {
      CoverageState _elvis = null;
      CoverageContext<PTraceable> _coverageContext = new CoverageContext<PTraceable>(traceable, scope);
      CoverageState _get = coverage.get(_coverageContext);
      if (_get != null) {
        _elvis = _get;
      } else {
        _elvis = CoverageState.NOT_REPRESENTED_UNKNOWN_REASON;
      }
      final CoverageState state = _elvis;
      final Function1<PTraceable, Boolean> _function = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          CoverageContext<PTraceable> _coverageContext = new CoverageContext<PTraceable>(it, scope);
          CoverageState _get = coverage.get(_coverageContext);
          return Boolean.valueOf((!Objects.equal(_get, CoverageState.NOT_COVERED)));
        }
      };
      final boolean childrenCovered = IterableExtensions.forall(children, _function);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("style=\"");
      String _style = CoverageReporter.getStyle(state, childrenCovered);
      _builder.append(_style, "");
      _builder.append("\" title=\"");
      String _title = CoverageReporter.getTitle(state, childrenCovered);
      _builder.append(_title, "");
      _builder.append("\"");
      _xblockexpression = _builder.toString();
    }
    return _xblockexpression;
  }
  
  private static String getLegend(final CoverageState state, final boolean childrenCovered) {
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("<span style=\"");
    String _style = CoverageReporter.getStyle(state, childrenCovered);
    _builder.append(_style, "");
    _builder.append("\">");
    String _title = CoverageReporter.getTitle(state, childrenCovered);
    _builder.append(_title, "");
    _builder.append("</span><br/>");
    return _builder.toString();
  }
  
  private static String getTitle(final CoverageState state, final boolean childrenCovered) {
    String _switchResult = null;
    if (state != null) {
      switch (state) {
        case COVERED:
          String _xifexpression = null;
          if (childrenCovered) {
            _xifexpression = "Covered";
          } else {
            _xifexpression = "Partially covered";
          }
          _switchResult = _xifexpression;
          break;
        case NOT_COVERED:
          _switchResult = "Not covered";
          break;
        case NOT_REPRESENTED:
          _switchResult = "Not represented";
          break;
        case NOT_REPRESENTED_UNKNOWN_REASON:
          _switchResult = "Not represented by error - please report";
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
  
  private static String getStyle(final CoverageState state, final boolean childrenCovered) {
    String _switchResult = null;
    if (state != null) {
      switch (state) {
        case COVERED:
          String _xifexpression = null;
          if (childrenCovered) {
            _xifexpression = "background:springgreen";
          } else {
            _xifexpression = "background:yellow";
          }
          _switchResult = _xifexpression;
          break;
        case NOT_COVERED:
          _switchResult = "background:tomato";
          break;
        case NOT_REPRESENTED:
          _switchResult = "color:lightgray";
          break;
        case NOT_REPRESENTED_UNKNOWN_REASON:
          _switchResult = "color:red";
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
}
