/**
 * Copyright (c) 2010-2017, Grill Balázs, IncQueryLabs
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Balázs - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.coverage;

import com.google.common.base.Objects;
import com.google.common.base.Supplier;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import org.eclipse.viatra.query.runtime.api.AdvancedViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.IPatternMatch;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.ViatraQueryMatcher;
import org.eclipse.viatra.query.runtime.api.scope.QueryScope;
import org.eclipse.viatra.query.runtime.emf.EMFScope;
import org.eclipse.viatra.query.runtime.exception.ViatraQueryException;
import org.eclipse.viatra.query.runtime.matchers.backend.IQueryBackend;
import org.eclipse.viatra.query.runtime.matchers.planning.QueryProcessingException;
import org.eclipse.viatra.query.runtime.matchers.planning.SubPlan;
import org.eclipse.viatra.query.runtime.matchers.psystem.PBody;
import org.eclipse.viatra.query.runtime.matchers.psystem.PConstraint;
import org.eclipse.viatra.query.runtime.matchers.psystem.PTraceable;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQueries;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQuery;
import org.eclipse.viatra.query.runtime.matchers.psystem.rewriters.IPTraceableTraceProvider;
import org.eclipse.viatra.query.runtime.rete.matcher.ReteBackendFactory;
import org.eclipse.viatra.query.runtime.rete.matcher.ReteEngine;
import org.eclipse.viatra.query.runtime.rete.network.Network;
import org.eclipse.viatra.query.runtime.rete.network.Node;
import org.eclipse.viatra.query.runtime.rete.traceability.CompiledQuery;
import org.eclipse.viatra.query.runtime.rete.traceability.CompiledSubPlan;
import org.eclipse.viatra.query.runtime.rete.traceability.RecipeTraceInfo;
import org.eclipse.viatra.query.testing.core.coverage.CoverageContext;
import org.eclipse.viatra.query.testing.core.coverage.CoverageInfo;
import org.eclipse.viatra.query.testing.core.coverage.CoverageState;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.Functions.Function2;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.MapExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure2;

/**
 * This utility class can determine trace connection between Nodes in a Rete network and elements in a PQuery model.
 * @since 1.6
 */
@SuppressWarnings("all")
public class ReteNetworkTrace {
  private final ReteEngine reteEngine;
  
  private final EMFScope scope;
  
  private final Multimap<PTraceable, Node> traceableToNodeMap = Multimaps.<PTraceable, Node>newSetMultimap(CollectionLiterals.<PTraceable, Collection<Node>>newHashMap(), new Supplier<Set<Node>>() {
    @Override
    public Set<Node> get() {
      return CollectionLiterals.<Node>newHashSet();
    }
  });
  
  private final IPTraceableTraceProvider traceProvider;
  
  private void updateMapWithCanonicalTraceable(final PTraceable derivedTraceable, final Node node, final IPTraceableTraceProvider traceProvider, final Multimap<PTraceable, Node> traceableToNodeMap) {
    Iterable<PTraceable> _canonicalTraceables = traceProvider.getCanonicalTraceables(derivedTraceable);
    final Procedure1<PTraceable> _function = new Procedure1<PTraceable>() {
      @Override
      public void apply(final PTraceable traceable) {
        traceableToNodeMap.put(traceable, node);
      }
    };
    IterableExtensions.<PTraceable>forEach(_canonicalTraceables, _function);
  }
  
  public ReteNetworkTrace(final ViatraQueryMatcher<?> matcher, final IPTraceableTraceProvider traceProvider) throws ViatraQueryException, QueryProcessingException {
    ViatraQueryEngine _engine = matcher.getEngine();
    ReteBackendFactory _reteBackendFactory = new ReteBackendFactory();
    IQueryBackend _queryBackend = ((AdvancedViatraQueryEngine) _engine).getQueryBackend(_reteBackendFactory);
    this.reteEngine = ((ReteEngine) _queryBackend);
    ViatraQueryEngine _engine_1 = matcher.getEngine();
    QueryScope _scope = _engine_1.getScope();
    this.scope = ((EMFScope) _scope);
    this.traceProvider = traceProvider;
    Network _reteNet = this.reteEngine.getReteNet();
    Set<RecipeTraceInfo> _recipeTraces = _reteNet.getRecipeTraces();
    final Procedure1<RecipeTraceInfo> _function = new Procedure1<RecipeTraceInfo>() {
      @Override
      public void apply(final RecipeTraceInfo recipeTrace) {
        final Node node = recipeTrace.getNode();
        boolean _matched = false;
        if (!_matched) {
          if (recipeTrace instanceof CompiledSubPlan) {
            _matched=true;
            SubPlan _subPlan = ((CompiledSubPlan)recipeTrace).getSubPlan();
            Set<PConstraint> _allEnforcedConstraints = _subPlan.getAllEnforcedConstraints();
            final Procedure1<PConstraint> _function = new Procedure1<PConstraint>() {
              @Override
              public void apply(final PConstraint it) {
                ReteNetworkTrace.this.updateMapWithCanonicalTraceable(it, node, traceProvider, ReteNetworkTrace.this.traceableToNodeMap);
              }
            };
            IterableExtensions.<PConstraint>forEach(_allEnforcedConstraints, _function);
          }
        }
        if (!_matched) {
          if (recipeTrace instanceof CompiledQuery) {
            _matched=true;
            PQuery _query = ((CompiledQuery)recipeTrace).getQuery();
            ReteNetworkTrace.this.updateMapWithCanonicalTraceable(_query, node, traceProvider, ReteNetworkTrace.this.traceableToNodeMap);
            Map<PBody, ? extends RecipeTraceInfo> _parentRecipeTracesPerBody = ((CompiledQuery)recipeTrace).getParentRecipeTracesPerBody();
            final Procedure2<PBody, RecipeTraceInfo> _function = new Procedure2<PBody, RecipeTraceInfo>() {
              @Override
              public void apply(final PBody derivedBody, final RecipeTraceInfo parentRecipeTrace) {
                Node _node = parentRecipeTrace.getNode();
                ReteNetworkTrace.this.updateMapWithCanonicalTraceable(derivedBody, _node, traceProvider, ReteNetworkTrace.this.traceableToNodeMap);
              }
            };
            MapExtensions.forEach(_parentRecipeTracesPerBody, _function);
          }
        }
      }
    };
    IterableExtensions.<RecipeTraceInfo>forEach(_recipeTraces, _function);
  }
  
  /**
   * Find all nodes in the Rete network which originate from the given {@link PTraceable}
   */
  public Iterable<Node> findNodes(final PTraceable traceable) {
    return this.traceableToNodeMap.get(traceable);
  }
  
  /**
   * Extract the coverage of a PQuery based on a Rete coverage
   */
  public CoverageInfo<PTraceable> traceCoverage(final PQuery pQuery, final CoverageInfo<Node> reteCoverage) {
    CoverageInfo<PTraceable> _xblockexpression = null;
    {
      final CoverageInfo<PTraceable> coverage = new CoverageInfo<PTraceable>();
      Set<PTraceable> _traceables = PQueries.getTraceables(pQuery);
      final Procedure1<PTraceable> _function = new Procedure1<PTraceable>() {
        @Override
        public void apply(final PTraceable traceable) {
          Iterable<Node> _findNodes = ReteNetworkTrace.this.findNodes(traceable);
          final Function1<Node, CoverageState> _function = new Function1<Node, CoverageState>() {
            @Override
            public CoverageState apply(final Node it) {
              CoverageContext<Node> _create = CoverageContext.<Node>create(it, ReteNetworkTrace.this.scope);
              return reteCoverage.get(_create);
            }
          };
          Iterable<CoverageState> _map = IterableExtensions.<Node, CoverageState>map(_findNodes, _function);
          final Function2<CoverageState, CoverageState, CoverageState> _function_1 = new Function2<CoverageState, CoverageState, CoverageState>() {
            @Override
            public CoverageState apply(final CoverageState r, final CoverageState t) {
              return r.best(t);
            }
          };
          final CoverageState coverageBasedOnTracedReteNodes = IterableExtensions.<CoverageState, CoverageState>fold(_map, CoverageState.NOT_REPRESENTED_UNKNOWN_REASON, _function_1);
          CoverageState _xifexpression = null;
          boolean _and = false;
          boolean _equals = Objects.equal(coverageBasedOnTracedReteNodes, CoverageState.NOT_REPRESENTED_UNKNOWN_REASON);
          if (!_equals) {
            _and = false;
          } else {
            boolean _hasRemovalReason = ReteNetworkTrace.this.hasRemovalReason(traceable);
            _and = _hasRemovalReason;
          }
          if (_and) {
            _xifexpression = CoverageState.NOT_REPRESENTED;
          } else {
            _xifexpression = coverageBasedOnTracedReteNodes;
          }
          final CoverageState amendedCoverage = _xifexpression;
          CoverageContext<PTraceable> _create = CoverageContext.<PTraceable>create(traceable, ReteNetworkTrace.this.scope);
          coverage.put(_create, amendedCoverage);
        }
      };
      IterableExtensions.<PTraceable>forEach(_traceables, _function);
      _xblockexpression = coverage;
    }
    return _xblockexpression;
  }
  
  private boolean hasRemovalReason(final PTraceable canonical) {
    return this.traceProvider.isRemoved(canonical);
  }
  
  /**
   * Extract the coverage of a ViatraQueryMatcher based on a Rete coverage
   */
  public CoverageInfo<PTraceable> traceCoverage(final ViatraQueryMatcher<?> matcher, final CoverageInfo<Node> reteCoverage) {
    IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>> _specification = matcher.getSpecification();
    PQuery _internalQueryRepresentation = _specification.getInternalQueryRepresentation();
    return this.traceCoverage(_internalQueryRepresentation, reteCoverage);
  }
}
